#! /usr/bin/env python
#
#    Implements the Quad-Edge model data structure, as described by
#       http://doi.acm.org/10.1145/282918.282923
#
#    Copyright (C) 2002  Michael Urman
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#

class Edge(object):
    """Store a single edge as part of the Quad-Edge data structure"""

    # make storage more efficient: struct like, rather than dictionary
    __slots__ = ('e0','e1','e2','e3', '_qnext', '_qprev', '_data', '_next')


    def __init__(self, org, dest):
        """Edge(self, org, dest) -> Edge
        
        Initialize an edge going from vertex /org/ to vertex /dest/"""

        # special case org == dest:
        if org is dest:
            self._qprev = org
            self._qnext = self
        else:
            self.e0 = self
            self.e1 = Edge(self, self)
            self.e2 = Edge(self, self)
            self.e3 = Edge(self, self)
            self._qprev = self.e3
            self._qnext = self.e0
            self.e0._next = self.e0
            self.e1._next = self.e3
            self.e2._next = self.e2
            self.e3._next = self.e1

            raise NotImplementedError("SetEndPoints???")
            SetEndPoints(org,dest)

    def onext(self): return self._next
    onext = property(onext, None, None, "Edge ONext")

    def sym(self): return self._qnext._qnext
    sym = property(sym, None, None, "Edge Sym")

    def rot(self): return self._qnext
    rot = property(rot, None, None, "Edge Rot")

    def invrot(self): return self._qprev
    invrot = property(invrot, None, None, "Edge Inverse Rot")

    def oprev(self):
        # return self.rot.onext.rot
        return self._qnext._next._qnext
    oprev = property(oprev, None, None, "Edge OPrev")

    def dnext(self):
        # return self.sym.onext.sym
        return self._qnext._qnext._next._qnext._qnext
    dnext = property(dnext, None, None, "Edge DNext")

    def dprev(self):
        # return self.invrot.onext.invrot
        return self._qprev._next._qprev
    dprev = property(dprev, None, None, "Edge DPrev")

    def lnext(self):
        # return self.invrot.onext.rot
        return self._qprev._next._qnext
    lnext = property(lnext, None, None, "Edge LNext")

    def lprev(self):
        # return self.onext.sym
        return self._next._qnext._qnext
    lprev = property(lprev, None, None, "Edge LPrev")

    def rnext(self):
        # return self.rot.onext.invrot
        return self._qnext._next._qprev
    rnext = property(rnext, None, None, "Edge RNext")

    def rprev(self):
        # return self.sym.onext
        return self._qnext._qnext._next
    rprev = property(rprev, None, None, "Edge RPrev")

    def get_org(self): return self._data
    def set_org(self, org): self._data = org; return org
    org = property(get_org, set_org, "Edge's Origin")

    def get_dest(self): return self._qnext._qnext._data
    def set_dest(self, dest): self._qnext._qnext = dest; return dest
    dest = property(get_dest, set_dest, "Edge's Dest")

    def splice(a, b):
        """splice(a, b)

        Splice edges a and b by swapping their next's and onext.rot.onext's"""

        a._next, b._next, a._next._qnext._next, b._next._qnext._next = \
            b._next, a._next, b._next._qnext._next, a.onext._qnext._next

    splice = classmethod(splice)
