#! /usr/bin/env python
#
#    Transform model - wrap a model in this to be able to move and size it
#    Copyright (C) 2002  Michael Urman
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#

import model, quaternion
import math

try:
    from OpenGL.GL import *
except:
    print 'transform requires pyopengl'
    raise SystemExit

class Transform(model.Model):
    """Transformation nodes for hierarchical models"""
    def __init__(self, translate=None, rotate=None, scale=None):
        model.Model.__init__(self)
        self._x_translate = translate or [0,0,0]
        self._x_rotate = rotate or quaternion.Quaternion((0, (1,0,0)))
        self._x_scale = scale or [1,1,1]

    def draw(self, *args, **kvargs):
        """draw(self, *args, **kvargs) -> None

        Applies transformations and calls children's draw() routine"""

        glPushMatrix()
        glTranslate(*self._x_translate)
        glRotatef(self._x_rotate.angle(), *self._x_rotate.axis())

        glScale(*self._x_scale)
        model.Model.draw(self, *args, **kvargs)
        glPopMatrix()


    def rotate(self, angle, axis):
        """rotate(self, angle, axis) -> None

        Rotate model by angle around the axis (x,y,z)"""

        # normalize axis
        try:
            al = 1/math.sqrt(axis[0]*axis[0]+axis[1]*axis[1]+axis[2]*axis[2])
        except ZeroDivisionError:
            return self._x_rotate
        axis = axis[0] * al, axis[1]*al, axis[2]*al

        self._x_rotate.rotate(angle, axis)

        return self._x_rotate

    def translate(self, delta):
        """translate(self, delta) -> None

        Translate model by delta (x,y,z)"""

        self._x_translate[0] += delta[0]
        self._x_translate[1] += delta[1]
        self._x_translate[2] += delta[2]
        return self._x_translate[:]

    def scale(self, factor):
        """scale(self, factor) -> None

        Scale model by factor (x,y,z)"""

        if factor: self._x_scale = [self._x_scale[i]*factor[i] for i in range(3)]
        scale = self._x_scale
        if scale[0]+scale[1]+scale[2] == 0: self._x_scale = [0.1, 0.1, 0.1]
        return self._x_scale[:]

